@@ -45,7 +45,6 @@ def laplace_smoothing(x, n_categories, eps = 1e-5):
 
 def sample_vectors(samples, num):
     num_samples, device = samples.shape[0], samples.device
-
     if num_samples >= num:
         indices = torch.randperm(num_samples, device = device)[:num]
     else:
@@ -53,6 +52,9 @@ def sample_vectors(samples, num):
 
     return samples[indices]
 
+def batched_sample_vectors(samples, num):
+    return torch.stack([sample_vectors(sample, num) for sample in samples.unbind(dim = 0)], dim = 0)
+
 def pad_shape(shape, size, dim = 0):
     return [size if i == dim else s for i, s in enumerate(shape)]
 
@@ -103,51 +105,76 @@ def sample_vectors_distributed(local_samples, num):
     distributed.broadcast(samples_per_rank, src = 0)
     samples_per_rank = samples_per_rank.tolist()
 
-    local_samples = sample_vectors(local_samples, samples_per_rank[rank])
+    local_samples = batched_sample_vectors(local_samples, samples_per_rank[rank])
     all_samples = all_gather_variably_sized(local_samples, samples_per_rank, dim = 0)
     return torch.cat(all_samples, dim = 0)
 
-def kmeans(samples, num_clusters, num_iters = 10, use_cosine_sim = False, 
-           sample_fn = sample_vectors, all_reduce_fn = noop):
-    dim, dtype, device = samples.shape[-1], samples.dtype, samples.device
+def batched_bincount(x, *, minlength):
+    batch, dtype, device = x.shape[0], x.dtype, x.device
+    target = torch.zeros(batch, minlength, dtype = dtype, device = device)
+    values = torch.ones_like(x)
+    target.scatter_add_(-1, x, values)
+    return target
+
+def kmeans(
+    samples,
+    num_clusters,
+    num_iters = 10,
+    use_cosine_sim = False,
+    sample_fn = batched_sample_vectors,
+    all_reduce_fn = noop
+):
+    num_codebooks, dim, dtype, device = samples.shape[0], samples.shape[-1], samples.dtype, samples.device
+
     means = sample_fn(samples, num_clusters)
 
     for _ in range(num_iters):
         if use_cosine_sim:
-            dists = samples @ means.t()
+            dists = samples @ rearrange(means, 'h n d -> h d n')
         else:
-            diffs = rearrange(samples, 'n d -> n () d') \
-                    - rearrange(means, 'c d -> () c d')
+            diffs = rearrange(samples, '... n d -> ... n 1 d') - \
+                    rearrange(means, '... c d -> ... 1 c d')
             dists = -(diffs ** 2).sum(dim = -1)
 
         buckets = torch.argmax(dists, dim = -1)
-        bins = torch.bincount(buckets, minlength = num_clusters)
+        bins = batched_bincount(buckets, minlength = num_clusters)
         all_reduce_fn(bins)
 
         zero_mask = bins == 0
         bins_min_clamped = bins.masked_fill(zero_mask, 1)
 
-        new_means = buckets.new_zeros(num_clusters, dim, dtype = dtype)
-        new_means.scatter_add_(0, repeat(buckets, 'n -> n d', d = dim), samples)
-        new_means = new_means / bins_min_clamped[..., None]
+        new_means = buckets.new_zeros(num_codebooks, num_clusters, dim, dtype = dtype)
+
+        new_means.scatter_add_(1, repeat(buckets, 'h n -> h n d', d = dim), samples)
+        new_means = new_means / rearrange(bins_min_clamped, '... -> ... 1')
         all_reduce_fn(new_means)
 
         if use_cosine_sim:
             new_means = l2norm(new_means)
 
-        means = torch.where(zero_mask[..., None], means, new_means)
+        means = torch.where(
+            rearrange(zero_mask, '... -> ... 1'),
+            means,
+            new_means
+        )
 
     return means, bins
 
+def batched_embedding(indices, embeds):
+    batch, dim = indices.shape[1], embeds.shape[-1]
+    indices = repeat(indices, 'h b n -> h b n d', d = dim)
+    embeds = repeat(embeds, 'h c d -> h b c d', b = batch)
+    return embeds.gather(2, indices)
+
 # regularization losses
 
 def orthgonal_loss_fn(t):
     # eq (2) from https://arxiv.org/abs/2112.00384
-    n = t.shape[0]
+    h, n = t.shape[:2]
     normed_codes = l2norm(t)
-    identity = torch.eye(n, device = t.device)
-    cosine_sim = einsum('i d, j d -> i j', normed_codes, normed_codes)
-    return ((cosine_sim - identity) ** 2).sum() / (n ** 2)
+    identity = repeat(torch.eye(n, device = t.device), 'i j -> h i j', h = h)
+    cosine_sim = einsum('h i d, h j d -> h i j', normed_codes, normed_codes)
+    return ((cosine_sim - identity) ** 2).sum() / (h * n ** 2)
 
 # distance types
 
@@ -156,6 +183,7 @@ class EuclideanCodebook(nn.Module):
         self,
         dim,
         codebook_size,
+        num_codebooks = 1,
         kmeans_init = False,
         kmeans_iters = 10,
         decay = 0.8,
@@ -168,19 +196,21 @@ class EuclideanCodebook(nn.Module):
         super().__init__()
         self.decay = decay
         init_fn = uniform_init if not kmeans_init else torch.zeros
-        embed = init_fn(codebook_size, dim)
+        embed = init_fn(num_codebooks, codebook_size, dim)
 
         self.codebook_size = codebook_size
+        self.num_codebooks = num_codebooks
+
         self.kmeans_iters = kmeans_iters
         self.eps = eps
         self.threshold_ema_dead_code = threshold_ema_dead_code
         self.sample_codebook_temp = sample_codebook_temp
 
-        self.sample_fn = sample_vectors_distributed if use_ddp else sample_vectors
+        self.sample_fn = sample_vectors_distributed if use_ddp else batched_sample_vectors
         self.all_reduce_fn = distributed.all_reduce if use_ddp else noop
 
         self.register_buffer('initted', torch.Tensor([not kmeans_init]))
-        self.register_buffer('cluster_size', torch.zeros(codebook_size))
+        self.register_buffer('cluster_size', torch.zeros(num_codebooks, codebook_size))
         self.register_buffer('embed_avg', embed.clone())
 
         self.learnable_codebook = learnable_codebook
@@ -194,65 +224,89 @@ class EuclideanCodebook(nn.Module):
         if self.initted:
             return
 
-        embed, cluster_size = kmeans(data, self.codebook_size, self.kmeans_iters,
-                                     sample_fn = self.sample_fn, all_reduce_fn = self.all_reduce_fn)
+        embed, cluster_size = kmeans(
+            data,
+            self.codebook_size,
+            self.kmeans_iters,
+            sample_fn = self.sample_fn,
+            all_reduce_fn = self.all_reduce_fn
+        )
 
         self.embed.data.copy_(embed)
         self.embed_avg.data.copy_(embed.clone())
         self.cluster_size.data.copy_(cluster_size)
         self.initted.data.copy_(torch.Tensor([True]))
 
-    def replace(self, samples, mask):
-        self.embed.data[mask] = self.sample_fn(samples, mask.sum().item())
+    def replace(self, batch_samples, batch_mask):
+        batch_samples = l2norm(batch_samples)
+
+        for ind, (samples, mask) in enumerate(zip(batch_samples.unbind(dim = 0), batch_mask.unbind(dim = 0))):
+            if not torch.any(mask):
+                continue
+
+            sampled = self.sample_fn(rearrange(samples, '... -> 1 ...'), mask.sum().item())
+            self.embed.data[ind][mask] = rearrange(sampled, '1 ... -> ...')
 
     def expire_codes_(self, batch_samples):
         if self.threshold_ema_dead_code == 0:
             return
 
         expired_codes = self.cluster_size < self.threshold_ema_dead_code
+
         if not torch.any(expired_codes):
             return
-        batch_samples = rearrange(batch_samples, '... d -> (...) d')
-        self.replace(batch_samples, mask = expired_codes)
+
+        batch_samples = rearrange(batch_samples, 'h ... d -> h (...) d')
+        self.replace(batch_samples, batch_mask = expired_codes)
 
     @autocast(enabled = False)
     def forward(self, x):
+        needs_codebook_dim = x.ndim < 4
+
         x = x.float()
 
+        if needs_codebook_dim:
+            x = rearrange(x, '... -> 1 ...')
+
         shape, dtype = x.shape, x.dtype
-        flatten = rearrange(x, '... d -> (...) d')
+        flatten = rearrange(x, 'h ... d -> h (...) d')
 
         self.init_embed_(flatten)
 
         embed = self.embed if not self.learnable_codebook else self.embed.detach()
-        embed = self.embed.t()
+
+        embed = rearrange(embed, '... n d -> ... d n')
 
         dist = -(
-            flatten.pow(2).sum(1, keepdim=True)
+            (flatten ** 2).sum(dim = -1, keepdim=True)
             - 2 * flatten @ embed
-            + embed.pow(2).sum(0, keepdim=True)
+            + (embed ** 2).sum(dim = -2, keepdim=True)
         )
 
         embed_ind = gumbel_sample(dist, dim = -1, temperature = self.sample_codebook_temp)
         embed_onehot = F.one_hot(embed_ind, self.codebook_size).type(dtype)
         embed_ind = embed_ind.view(*shape[:-1])
-        quantize = F.embedding(embed_ind, self.embed)
+
+        quantize = batched_embedding(embed_ind, self.embed)
 
         if self.training:
-            cluster_size = embed_onehot.sum(0)
-            self.all_reduce_fn(cluster_size)
+            cluster_size = embed_onehot.sum(dim = 1)
 
+            self.all_reduce_fn(cluster_size)
             ema_inplace(self.cluster_size, cluster_size, self.decay)
 
-            embed_sum = flatten.t() @ embed_onehot
+            embed_sum = einsum('h n d, h n c -> h c d', flatten, embed_onehot)
             self.all_reduce_fn(embed_sum)
 
-            ema_inplace(self.embed_avg, embed_sum.t(), self.decay)
             cluster_size = laplace_smoothing(self.cluster_size, self.codebook_size, self.eps) * self.cluster_size.sum()
-            embed_normalized = self.embed_avg / cluster_size.unsqueeze(1)
+
+            embed_normalized = self.embed_avg / rearrange(cluster_size, '... -> ... 1')
             self.embed.data.copy_(embed_normalized)
             self.expire_codes_(x)
 
+        if needs_codebook_dim:
+            quantize, embed_ind = map(lambda t: rearrange(t, '1 ... -> ...'), (quantize, embed_ind))
+
         return quantize, embed_ind
 
 class CosineSimCodebook(nn.Module):
@@ -260,6 +314,7 @@ class CosineSimCodebook(nn.Module):
         self,
         dim,
         codebook_size,
+        num_codebooks = 1,
         kmeans_init = False,
         kmeans_iters = 10,
         decay = 0.8,
@@ -273,21 +328,23 @@ class CosineSimCodebook(nn.Module):
         self.decay = decay
 
         if not kmeans_init:
-            embed = l2norm(uniform_init(codebook_size, dim))
+            embed = l2norm(uniform_init(num_codebooks, codebook_size, dim))
         else:
-            embed = torch.zeros(codebook_size, dim)
+            embed = torch.zeros(num_codebooks, codebook_size, dim)
 
         self.codebook_size = codebook_size
+        self.num_codebooks = num_codebooks
+
         self.kmeans_iters = kmeans_iters
         self.eps = eps
         self.threshold_ema_dead_code = threshold_ema_dead_code
         self.sample_codebook_temp = sample_codebook_temp
 
-        self.sample_fn = sample_vectors_distributed if use_ddp else sample_vectors
+        self.sample_fn = sample_vectors_distributed if use_ddp else batched_sample_vectors
         self.all_reduce_fn = distributed.all_reduce if use_ddp else noop
 
         self.register_buffer('initted', torch.Tensor([not kmeans_init]))
-        self.register_buffer('cluster_size', torch.zeros(codebook_size))
+        self.register_buffer('cluster_size', torch.zeros(num_codebooks, codebook_size))
 
         self.learnable_codebook = learnable_codebook
         if learnable_codebook:
@@ -300,34 +357,53 @@ class CosineSimCodebook(nn.Module):
         if self.initted:
             return
 
-        embed, cluster_size = kmeans(data, self.codebook_size, self.kmeans_iters, use_cosine_sim = True,
-                                     sample_fn = self.sample_fn, all_reduce_fn = self.all_reduce_fn)
+        embed, cluster_size = kmeans(
+            data,
+            self.codebook_size,
+            self.kmeans_iters,
+            use_cosine_sim = True,
+            sample_fn = self.sample_fn,
+            all_reduce_fn = self.all_reduce_fn
+        )
 
         self.embed.data.copy_(embed)
         self.cluster_size.data.copy_(cluster_size)
         self.initted.data.copy_(torch.Tensor([True]))
 
-    def replace(self, samples, mask):
-        samples = l2norm(samples)
-        self.embed.data[mask] = self.sample_fn(samples, mask.sum().item())
+    def replace(self, batch_samples, batch_mask):
+        batch_samples = l2norm(batch_samples)
+
+        for ind, (samples, mask) in enumerate(zip(batch_samples.unbind(dim = 0), batch_mask.unbind(dim = 0))):
+            if not torch.any(mask):
+                continue
+
+            sampled = self.sample_fn(rearrange(samples, '... -> 1 ...'), mask.sum().item())
+            self.embed.data[ind][mask] = rearrange(sampled, '1 ... -> ...')
 
     def expire_codes_(self, batch_samples):
         if self.threshold_ema_dead_code == 0:
             return
 
         expired_codes = self.cluster_size < self.threshold_ema_dead_code
+
         if not torch.any(expired_codes):
             return
-        batch_samples = rearrange(batch_samples, '... d -> (...) d')
-        self.replace(batch_samples, mask = expired_codes)
+
+        batch_samples = rearrange(batch_samples, 'h ... d -> h (...) d')
+        self.replace(batch_samples, batch_mask = expired_codes)
 
     @autocast(enabled = False)
     def forward(self, x):
+        needs_codebook_dim = x.ndim < 4
+
         x = x.float()
 
+        if needs_codebook_dim:
+            x = rearrange(x, '... -> 1 ...')
+
         shape, dtype = x.shape, x.dtype
 
-        flatten = rearrange(x, '... d -> (...) d')
+        flatten = rearrange(x, 'h ... d -> h (...) d')
         flatten = l2norm(flatten)
 
         self.init_embed_(flatten)
@@ -335,15 +411,15 @@ class CosineSimCodebook(nn.Module):
         embed = self.embed if not self.learnable_codebook else self.embed.detach()
         embed = l2norm(embed)
 
-        dist = flatten @ embed.t()
+        dist = einsum('h n d, h c d -> h n c', flatten, embed)
         embed_ind = gumbel_sample(dist, dim = -1, temperature = self.sample_codebook_temp)
         embed_onehot = F.one_hot(embed_ind, self.codebook_size).type(dtype)
         embed_ind = embed_ind.view(*shape[:-1])
 
-        quantize = F.embedding(embed_ind, self.embed)
+        quantize = batched_embedding(embed_ind, self.embed)
 
         if self.training:
-            bins = embed_onehot.sum(0)
+            bins = embed_onehot.sum(dim = 1)
             self.all_reduce_fn(bins)
 
             ema_inplace(self.cluster_size, bins, self.decay)
@@ -351,16 +427,24 @@ class CosineSimCodebook(nn.Module):
             zero_mask = (bins == 0)
             bins = bins.masked_fill(zero_mask, 1.)
 
-            embed_sum = flatten.t() @ embed_onehot
+            embed_sum = einsum('h n d, h n c -> h c d', flatten, embed_onehot)
             self.all_reduce_fn(embed_sum)
 
-            embed_normalized = (embed_sum / bins.unsqueeze(0)).t()
+            embed_normalized = embed_sum / rearrange(bins, '... -> ... 1')
             embed_normalized = l2norm(embed_normalized)
-            embed_normalized = torch.where(zero_mask[..., None], embed,
-                                           embed_normalized)
+
+            embed_normalized = torch.where(
+                rearrange(zero_mask, '... -> ... 1'),
+                embed,
+                embed_normalized
+            )
+
             ema_inplace(self.embed, embed_normalized, self.decay)
             self.expire_codes_(x)
 
+        if needs_codebook_dim:
+            quantize, embed_ind = map(lambda t: rearrange(t, '1 ... -> ...'), (quantize, embed_ind))
+
         return quantize, embed_ind
 
 # main class
@@ -372,6 +456,7 @@ class VectorQuantize(nn.Module):
         codebook_size,
         codebook_dim = None,
         heads = 1,
+        separate_codebook_per_head = False,
         decay = 0.8,
         eps = 1e-5,
         kmeans_init = False,
@@ -389,6 +474,8 @@ class VectorQuantize(nn.Module):
     ):
         super().__init__()
         self.heads = heads
+        self.separate_codebook_per_head = separate_codebook_per_head
+
         codebook_dim = default(codebook_dim, dim)
         codebook_input_dim = codebook_dim * heads
 
@@ -408,6 +495,7 @@ class VectorQuantize(nn.Module):
 
         self._codebook = codebook_class(
             dim = codebook_dim,
+            num_codebooks = heads if separate_codebook_per_head else 1,
             codebook_size = codebook_size,
             kmeans_init = kmeans_init,
             kmeans_iters = kmeans_iters,
@@ -443,7 +531,8 @@ class VectorQuantize(nn.Module):
         x = self.project_in(x)
 
         if is_multiheaded:
-            x = rearrange(x, 'b n (h d) -> (b h) n d', h = heads)
+            ein_rhs_eq = 'h b n d' if self.separate_codebook_per_head else '1 (b h) n d'
+            x = rearrange(x, f'b n (h d) -> {ein_rhs_eq}', h = heads)
 
         quantize, embed_ind = self._codebook(x)
 
@@ -474,8 +563,12 @@ class VectorQuantize(nn.Module):
                 loss = loss + orthogonal_reg_loss * self.orthogonal_reg_weight
 
         if is_multiheaded:
-            quantize = rearrange(quantize, '(b h) n d -> b n (h d)', h = heads)
-            embed_ind = rearrange(embed_ind, '(b h) n -> b n h', h = heads)
+            if self.separate_codebook_per_head:
+                quantize = rearrange(quantize, 'h b n d -> b n (h d)', h = heads)
+                embed_ind = rearrange(embed_ind, 'h b n -> b n h', h = heads)
+            else:
+                quantize = rearrange(quantize, '1 (b h) n d -> b n (h d)', h = heads)
+                embed_ind = rearrange(embed_ind, '1 (b h) n -> b n h', h = heads)
 
         quantize = self.project_out(quantize)
 
